//---------------------------------------------------------------------------
// DensitySeedSurvival
//---------------------------------------------------------------------------
#if !defined(DensitySeedSurvival_H)
  #define DensitySeedSurvival_H

#include "BehaviorBase.h"

class clGridBase;
class clSimManager;

/**
* Density-Dependent Seed Survival - version 1.0
*
* This behavior reduces the number of seeds due to density-dependent effects.
*
* The number of seeds this behavior leaves of a given species in a particular
* "Dispersed Seeds" grid cell is:
* @htmlonly
  <br><center>R<sub>sp</sub> = S<sub>sp</sub>*exp(-c*Den<sub>sp</sub><sup>&delta;</sup>)</center><br>
  @endhtmlonly
  where:</p>
* <ul>
* <li>R<sub>sp</sub> is the number of seedlings of a given species in the seed
* grid cell (competition is conspecific only)</li>
* <li>S<sub>sp</sub> is the potential number of seedlings of the species, i.e.
* the number of seeds dispersed to that grid cell</li>
* <li>Den<sub>sp</sub> is the density of seed of the species per square
* meter</li>
* <li>c is an input parameter - density-dependent slope</li>
* <li>@htmlonly &delta; @endhtmlonly is an input parameter - density-dependent
* steepness</li>
* </ul>
*
* Single seeds always survive.
*
* A fatal error is thrown if a disperse behavior is not also defined for the
* run.
*
* This class's namestring and parameter call string are both
* "Density Dependent Seed Survival".
* Copyright 2005 Charles D. Canham.
* @author Lora E. Murphy
*
* <br>Edit history:
* <br>-----------------
* <br>April 12, 2005 - Created (LEM)
* <br>November 15, 2006 - Fixed parameters to be area-independent and made sure
* that single seeds always survive (LEM)
*/
class clDensitySeedSurvival : virtual public clBehaviorBase {
//note: need the virtual keyword to avoid base class ambiguity.

 public:

  /**
  * Constructor.
  *
  * @param p_oSimManager Sim Manager object.
  */
  clDensitySeedSurvival(clSimManager *p_oSimManager);

  /**
  * Destructor.
  */
  ~clDensitySeedSurvival();

  /**
  * Reads in values from the parameter file.
  *
  * @param p_oDoc DOM tree of parsed input file.
  */
  void GetData(xercesc::DOMDocument *p_oDoc);

  /**
  * Performs density-dependent seed survival.  For each grid cell in the seed
  * grid, for each species to which this behavior applies, the number of seeds
  * left is calculated by assessing the equation above. A random round is used
  * to take care of fractional parts of seeds.
  */
  void Action();

  protected:

  /**Pointer to the "Dispersed Seeds" grid created by disperse behaviors.*/
  clGridBase *mp_oSeedGrid;

  /**Density-dependence steepness parameter.  Array size is total number of
  * species.*/
  float *mp_fDensDepSteepness;

  /**Density-dependence slope parameter.  Array size is total number of
  * species.*/
  float *mp_fDensDepSlope;

  /**Data member codes for seed grid for number of seeds.  Array size is #
  * total species.*/
  short int *mp_iSeedGridCode;

  /**Area of the "Dispersed Seeds" grid cells, in square meters.*/
  float m_fNormalCellArea;

  /**Area of the last row of "Dispersed Seeds" grid cells in the X direction,
  * in square meters.  This allows for grid cells that don't divide evenly
  * into plot lengths.*/
  float m_fXEdgeCellArea;

  /**Area of the last row of "Dispersed Seeds" grid cells in the Y direction,
  * in square meters.  This allows for grid cells that don't divide evenly
  * into plot lengths.*/
  float m_fYEdgeCellArea;

  /**Area of the last cell of "Dispersed Seeds" grid cell in both the X and Y
  * direction, in square meters.  This allows for grid cells that don't divide
  * evenly into plot lengths.*/
  float m_fXYEdgeCellArea;
};
//---------------------------------------------------------------------------
#endif // Germination_H

