//---------------------------------------------------------------------------
#include "Disturbance.h"
#include "SimManager.h"
#include "ParsingFunctions.h"
#include "GridBase.h"
#include "TreePopulation.h"
#include "Plot.h"
#include <stdio.h>
#include <fstream>

short int clDisturbance::m_iNumAllowedCutRanges = 4;

/////////////////////////////////////////////////////////////////////////////
// Constructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
clDisturbance::clDisturbance( clSimManager * p_oSimManager ) : clWorkerBase( p_oSimManager ), clBehaviorBase( p_oSimManager )
{
  try
  {
    strcpy( m_cNameString, "harvest" );

    //Versions
    m_fVersionNumber = 1;
    m_fMinimumVersionNumber = 1;

    //Null pointers, initialize variables to empty
    mp_oMasterCutsGrid = NULL;
    mp_oCutEventsGrid = NULL;
    mp_oResultsGrid = NULL;
    mp_iSpeciesCodes = NULL;
    mp_iRangeMinCodes = NULL;
    mp_iRangeMaxCodes = NULL;
    mp_iRangeAmountCodes = NULL;
    mp_iDenCutCodes = NULL;
    mp_iBaCutCodes = NULL;

    //Allowed file types
    m_iNumAllowedTypes = 2;
    mp_iAllowedFileTypes = new int[m_iNumAllowedTypes];
    mp_iAllowedFileTypes[0] = parfile;
    mp_iAllowedFileTypes[1] = detailed_output;

    //Number of allowed cut ranges
    m_iNumAllowedCutRanges = 4;
  }
  catch ( modelErr & err )
  {
    throw( err );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::clDisturbance" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

/////////////////////////////////////////////////////////////////////////////
// Destructor
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
/////////////////////////////////////////////////////////////////////////////
clDisturbance::~clDisturbance()
{
  short int i;
  delete[] mp_iSpeciesCodes;
  delete[] mp_iRangeMinCodes;
  delete[] mp_iRangeMaxCodes;
  delete[] mp_iRangeAmountCodes;
  if ( mp_iDenCutCodes )
    for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
    {
      delete[] mp_iDenCutCodes[i];
      delete[] mp_iBaCutCodes[i];
    }
  delete[] mp_iDenCutCodes;
  delete[] mp_iBaCutCodes;
}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// SetNameData()
// Edit history:
// -------------
// May 11, 2004 - Created (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::SetNameData( char * cNameString )
{
  try
  {

    //Check the string passed and set the flags accordingly
    if ( strcmp( "harvest", cNameString ) == 0 )
    {
      m_iReasonCode = harvest;
      m_bIsHarvest = true;
    }
    else if ( strcmp( "episodic mortality", cNameString ) == 0 )
    {
      m_iReasonCode = disease;
      m_bIsHarvest = false;
    }
    else
    {
      //Throw an error
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      sprintf( stcErr.cMoreInfo, "%s%s%s", "Unrecognized namestring \"", cNameString, "\"." );
      strcpy( stcErr.cFunction, "clDisturbance::SetNameData" );
      throw( stcErr );
    }

  }
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::SetNameData" );
    throw( stcErr );
  }
}

//---------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// ReadHarvestParameterFileData()
// Edit history:
// -------------
// May 11, 2004 - Created (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::ReadHarvestParameterFileData( xercesc::DOMDocument * p_oDoc )
{
  clPackage * p_oOldPackage = NULL, //for finding where to place a new package
       * p_oNewPackage = NULL; //a newly created package
  DOMNodeList * p_oCutEventsList, //list of cut events
       * p_oCutSpecificsList; //sub-cut-event - species, cut range, cell
  DOMNode * p_oNode;
  DOMElement * p_oElement, * p_oChildElement;
  XMLCh *sVal;
  float fTemp, //for extracting float data from parameter file
       fMaxVal, //maximum cut range value
       * p_fRanges; //for sorting cut ranges
  int iNumHarvestEvents, //total number of cut events to extract
       iX, iY, //X and Y grid values for package
       iNumberTimesteps = mp_oSimManager->GetNumberOfTimesteps(), iTimestep, //timestep of harvest cut event
       iMinIndex, //index of the smallest value on a list
       iNumChildren, //number of sub-cut-events
       iCutType, //cut type - value matches enum cutType
       iTemp; //for extracting integer data from parameter file
  short int i, j, k; //loop counters
  char *cData,
       cTemp[100]; //for extracting text data

  //See if there's any harvest data in the parameter file
  sVal = XMLString::transcode( "harvest" );
  p_oCutEventsList = p_oDoc->getElementsByTagName( sVal );
  XMLString::release(&sVal);
  if ( 0 == p_oCutEventsList->getLength() ) return;

  //Now get the list of all cut events
  p_oNode = p_oCutEventsList->item( 0 );
  p_oElement = ( DOMElement * ) p_oNode;
  sVal = XMLString::transcode( "ha_cutEvent" );
  p_oCutEventsList = p_oElement->getElementsByTagName( sVal );
  XMLString::release(&sVal);

  //If there's no cut events, don't do anything else
  iNumHarvestEvents = p_oCutEventsList->getLength();
  if ( 0 == iNumHarvestEvents ) return;

  //*******************************************
  //Go through each cut event and add it as packages to the grids
  //*******************************************
  for ( i = 0; i < iNumHarvestEvents; i++ )
  {
    p_oNode = p_oCutEventsList->item( i );
    p_oElement = ( DOMElement * ) p_oNode;

    //Timestep first
    FillSingleValue( p_oElement, "ha_timestep", & iTimestep, true );

    //Validate it - needs to be between 0 and number of timesteps
    if ( iTimestep <= 0 || iTimestep > iNumberTimesteps )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clDisturbance::ReadHarvestParameterFileData" );
      sprintf( stcErr.cMoreInfo, "%s%d", "Invalid harvest cut timestep ", iTimestep );
      throw( stcErr );
    }

    //Add a package to the master cut events grid - they're in timestep order,
    //so keep it that way
    p_oOldPackage = NULL;
    p_oNewPackage = NULL;
    p_oOldPackage = mp_oMasterCutsGrid->GetFirstPackageOfCell( 0, 0 );
    while ( p_oOldPackage )
    {
      //Get this package's timestep
      p_oOldPackage->GetValue( m_iMasterTimestepCode, & iTemp );

      if ( iTemp >= iTimestep ) //here's where we'll drop it
             break;

      p_oNewPackage = p_oOldPackage; //use temporarily as a holder
      p_oOldPackage = p_oOldPackage->GetNextPackage();
    }

    //Create the new package
    if ( !p_oNewPackage ) //either there's no packages in grid, or new is first
           p_oNewPackage = mp_oMasterCutsGrid->CreatePackageOfCell( 0, 0 );
    else //make the new one after the one in p_oNewPackage
           p_oNewPackage = mp_oMasterCutsGrid->CreatePackage( p_oNewPackage );

    //Load up the new package with the timestep value
    p_oNewPackage->SetValue( m_iMasterTimestepCode, iTimestep );

    //Get species values and assign them
    sVal = XMLString::transcode( "ha_applyToSpecies" );
    p_oCutSpecificsList = p_oElement->getElementsByTagName( sVal );
    XMLString::release(&sVal);
    iNumChildren = p_oCutSpecificsList->getLength();
    for ( j = 0; j < iNumChildren; j++ )
    {
      p_oNode = p_oCutSpecificsList->item( j );
      p_oChildElement = ( DOMElement * ) p_oNode;
      iTemp = GetNodeSpeciesCode( p_oChildElement, mp_oPop );
      if ( -1 == iTemp )
      { //throw an error
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clDisturbance::ReadHarvestParameterFileData" );
        strcpy( stcErr.cMoreInfo, "Unrecognized species in harvest." );
        throw( stcErr );
      }
      p_oNewPackage->SetValue( mp_iSpeciesCodes[iTemp], true );
    }

    //Get the cut type
    FillSingleValue( p_oElement, "ha_cutType", cTemp, 100, true );
    //See if it's a value we recognize
    if ( strcmp( cTemp, "gap" ) == 0 ) iCutType = gap;
    else if ( strcmp( cTemp, "partial" ) == 0 ) iCutType = partial;
    else if ( strcmp( cTemp, "clear" ) == 0 ) iCutType = clear;
    else
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clDisturbance::ReadHarvestParameterFileData" );
      strcpy( stcErr.cMoreInfo, "Unrecognized value in ha_cutType." );
      throw( stcErr );
    }
    p_oNewPackage->SetValue( m_iCutTypeCode, iCutType );

    //Get the amount type if this is a partial cut
    if ( partial == iCutType )
    {
      FillSingleValue( p_oElement, "ha_cutAmountType", cTemp, 100, true );
      //See if it's a value we recognize
      if ( strcmp( cTemp, "percent of basal area" ) == 0 )
        p_oNewPackage->SetValue( m_iAmountTypeCode, percentBA );
      else if ( strcmp( cTemp, "percent of density" ) == 0 )
        p_oNewPackage->SetValue( m_iAmountTypeCode, percentDen );
      else if ( strcmp( cTemp, "absolute basal area" ) == 0 )
        p_oNewPackage->SetValue( m_iAmountTypeCode, absBA );
      else if ( strcmp( cTemp, "absolute density" ) == 0 )
        p_oNewPackage->SetValue( m_iAmountTypeCode, absDen );
      else
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clDisturbance::ReadHarvestParameterFileData" );
        strcpy( stcErr.cMoreInfo, "Unrecognized value in ha_cutAmountType." );
        throw( stcErr );
      }
    }
    else
    {
      //Set the amount type to -1
      iTemp = -1;
      p_oNewPackage->SetValue( m_iAmountTypeCode, iTemp );
    }

    //Get the cut ranges
    sVal = XMLString::transcode( "ha_dbhRange" );
    p_oCutSpecificsList = p_oElement->getElementsByTagName( sVal );
    XMLString::release(&sVal);

    iNumChildren = p_oCutSpecificsList->getLength();
    //If there are more ranges than allowed, throw an error
    if ( iNumChildren > m_iNumAllowedCutRanges )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clDisturbance::ReadHarvestParameterFileData" );
      strcpy( stcErr.cMoreInfo, "Too many dbh cut ranges." );
      throw( stcErr );
    }
    //Or, if there aren't any and this is a partial cut, throw an error
    if ( partial == iCutType && 0 == iNumChildren )
    {
      modelErr stcErr;
      stcErr.iErrorCode = DATA_MISSING;
      strcpy( stcErr.cFunction, "clDisturbance::ReadHarvestParameterFileData" );
      strcpy( stcErr.cMoreInfo, "ha_dbhRange" );
      throw( stcErr );
    }

    //Get the minimum dbh in each cut range
    p_fRanges = new float[iNumChildren];
    for ( j = 0; j < iNumChildren; j++ )
    {
      p_oNode = p_oCutSpecificsList->item( j );
      p_oChildElement = ( DOMElement * ) p_oNode;
      FillSingleValue( p_oChildElement, "ha_low", & fTemp, true );
      p_fRanges[j] = fTemp;

      //If the value is negative, throw an error
      if ( fTemp < 0 )
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clDisturbance::ReadHarvestParameterFileData" );
        strcpy( stcErr.cMoreInfo, "Harvest cut range value cannot be negative." );
        throw( stcErr );
      }
    }

    //Sort the values and place them in the package
    fMaxVal = 0;
    for ( j = 0; j < iNumChildren; j++ )
      if ( fMaxVal < p_fRanges[j] )
        fMaxVal = p_fRanges[j];

    fMaxVal += 1;

    for ( j = 0; j < iNumChildren; j++ )
    {
      fTemp = fMaxVal;
      for ( k = 0; k < iNumChildren; k++ )
        if ( p_fRanges[k] < fTemp )
        {
          iMinIndex = k;
          fTemp = p_fRanges[k];
        }

      p_oNode = p_oCutSpecificsList->item( iMinIndex );
      p_oChildElement = ( DOMElement * ) p_oNode;

      p_oNewPackage->SetValue( mp_iRangeMinCodes[j], p_fRanges[iMinIndex] );
      FillSingleValue( p_oChildElement, "ha_high", & fTemp, true );
      p_oNewPackage->SetValue( mp_iRangeMaxCodes[j], fTemp );
      FillSingleValue( p_oChildElement, "ha_amountToCut", & fTemp, true );
      p_oNewPackage->SetValue( mp_iRangeAmountCodes[j], fTemp );

      p_fRanges[iMinIndex] = fMaxVal;
    }

    //Create an ID number - just use the loop counter
    p_oNewPackage->SetValue( m_iMasterIDCode, ( int )i );

    //*******************************************
    //Now:  Get the list of cells to which to apply this cut and make a
    //package for each
    //*******************************************
    sVal = XMLString::transcode( "ha_applyToCell" );
    p_oCutSpecificsList = p_oElement->getElementsByTagName( sVal );
    XMLString::release(&sVal);
    iNumChildren = p_oCutSpecificsList->getLength();
    for ( j = 0; j < iNumChildren; j++ )
    {
      p_oNode = p_oCutSpecificsList->item( j );
      p_oChildElement = ( DOMElement * ) p_oNode;
      sVal = XMLString::transcode( "x" );
      cData = XMLString::transcode(p_oChildElement->getAttributeNode(sVal)->getNodeValue());  
      iX = atoi( cData );
      delete[] cData; cData = NULL;
      XMLString::release(&sVal);
      sVal = XMLString::transcode( "y" );
      cData = XMLString::transcode(p_oChildElement->getAttributeNode(sVal)->getNodeValue());
      iY = atoi( cData );
      XMLString::release(&sVal);
      delete[] cData; cData = NULL;

      p_oOldPackage = NULL;
      p_oNewPackage = NULL;
      p_oOldPackage = mp_oCutEventsGrid->GetFirstPackageOfCell( iX, iY );
      while ( p_oOldPackage )
      {
        //Get this package's timestep
        p_oOldPackage->GetValue( m_iCutTimestepCode, & iTemp );

        if ( iTemp >= iTimestep ) //here's where we'll drop it
               break;

        p_oNewPackage = p_oOldPackage; //use temporarily as a holder
        p_oOldPackage = p_oOldPackage->GetNextPackage();
      }

      //Create the new package
      if ( !p_oNewPackage ) //either there's no packages in grid, or new is first
             p_oNewPackage = mp_oCutEventsGrid->CreatePackageOfCell( iX, iY );
      else //make the new one after the one in p_oNewPackage
             p_oNewPackage = mp_oCutEventsGrid->CreatePackage( p_oNewPackage );

      //Load up the new package with the timestep and id values
      p_oNewPackage->SetValue( m_iCutTimestepCode, iTimestep );
      p_oNewPackage->SetValue( m_iCutIDCode, ( int )i );
    } //end of for (j = 0; j < iNumChildren; j++)

    delete[] p_fRanges;
  } //end of for (i = 0; i < iNumHarvestEvents; i++)

}


//////////////////////////////////////////////////////////////////////////////
// ReadMortEpParameterFileData()
// Edit history:
// -------------
// May 11, 2004 - Created (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::ReadMortEpParameterFileData( xercesc::DOMDocument * p_oDoc )
{
  clPackage * p_oOldPackage = NULL, //for finding where to place a new package
       * p_oNewPackage = NULL; //a newly created package
  DOMNodeList * p_oCutEventsList, //list of cut events
       * p_oCutSpecificsList; //sub-cut-event - species, cut range, cell
  DOMNode * p_oNode;
  DOMElement * p_oElement, * p_oChildElement;
  XMLCh *sVal;
  float fTemp, //for extracting float data from parameter file
       fMaxVal, //maximum cut range value
       * p_fRanges; //for sorting cut ranges
  int iNumDisturbanceEvents, //total number of cut events to extract
       iX, iY, //X and Y grid values for package
       iNumberTimesteps = mp_oSimManager->GetNumberOfTimesteps(), iTimestep, //timestep of mortality episode
       iMinIndex, //index of the smallest value on a list
       iNumChildren, //number of sub-cut-events
       iTemp; //for extracting integer data from parameter file
  short int i, j, k; //loop counters
  char cTemp[100], //for extracting text data
       *cData;

  //See if there's any mortality episode data in the parameter file
  sVal = XMLString::transcode( "disturbance" );
  p_oCutEventsList = p_oDoc->getElementsByTagName( sVal );
  XMLString::release(&sVal);
  if ( 0 == p_oCutEventsList->getLength() ) return;

  //Now get the list of all cut events
  p_oNode = p_oCutEventsList->item( 0 );
  p_oElement = ( DOMElement * ) p_oNode;
  sVal = XMLString::transcode( "ds_deathEvent" );
  p_oCutEventsList = p_oElement->getElementsByTagName( sVal );
  XMLString::release(&sVal);

  //If there's no mortality episodes, don't do anything else
  iNumDisturbanceEvents = p_oCutEventsList->getLength();
  if ( 0 == iNumDisturbanceEvents ) return;

  //*******************************************
  //Go through each mortality episode and add it as packages to the grids
  //*******************************************
  for ( i = 0; i < iNumDisturbanceEvents; i++ )
  {
    p_oNode = p_oCutEventsList->item( i );
    p_oElement = ( DOMElement * ) p_oNode;

    //Timestep first
    FillSingleValue( p_oElement, "ds_timestep", & iTimestep, true );

    //Validate it - needs to be between 0 and number of timesteps
    if ( iTimestep <= 0 || iTimestep > iNumberTimesteps )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clDisturbance::ReadMortEpParameterFileData" );
      sprintf( stcErr.cMoreInfo, "%s%d", "Invalid mortality episode cut timestep ", iTimestep );
      throw( stcErr );
    }

    //Add a package to the master cut events grid - they're in timestep order,
    //so keep it that way
    p_oOldPackage = NULL;
    p_oNewPackage = NULL;
    p_oOldPackage = mp_oMasterCutsGrid->GetFirstPackageOfCell( 0, 0 );
    while ( p_oOldPackage )
    {
      //Get this package's timestep
      p_oOldPackage->GetValue( m_iMasterTimestepCode, & iTemp );

      if ( iTemp >= iTimestep ) //here's where we'll drop it
             break;

      p_oNewPackage = p_oOldPackage; //use temporarily as a holder
      p_oOldPackage = p_oOldPackage->GetNextPackage();
    }

    //Create the new package
    if ( !p_oNewPackage ) //either there's no packages in grid, or new is first
           p_oNewPackage = mp_oMasterCutsGrid->CreatePackageOfCell( 0, 0 );
    else //make the new one after the one in p_oNewPackage
           p_oNewPackage = mp_oMasterCutsGrid->CreatePackage( p_oNewPackage );

    //Load up the new package with the timestep value
    p_oNewPackage->SetValue( m_iMasterTimestepCode, iTimestep );

    //Get species values and assign them
    sVal = XMLString::transcode( "ds_applyToSpecies" );
    p_oCutSpecificsList = p_oElement->getElementsByTagName( sVal );
    XMLString::release(&sVal);
    iNumChildren = p_oCutSpecificsList->getLength();
    for ( j = 0; j < iNumChildren; j++ )
    {
      p_oNode = p_oCutSpecificsList->item( j );
      p_oChildElement = ( DOMElement * ) p_oNode;
      iTemp = GetNodeSpeciesCode( p_oChildElement, mp_oPop );
      if ( -1 == iTemp )
      { //throw an error
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clDisturbance::ReadMortEpParameterFileData" );
        strcpy( stcErr.cMoreInfo, "Unrecognized species in mortality episode." );
        throw( stcErr );
      }
      p_oNewPackage->SetValue( mp_iSpeciesCodes[iTemp], true );
    }

    //Get the amount type
    FillSingleValue( p_oElement, "ds_cutAmountType", cTemp, 100, true );
    //See if it's a value we recognize
    if ( strcmp( cTemp, "percent of basal area" ) == 0 )
      p_oNewPackage->SetValue( m_iAmountTypeCode, percentBA );
    else if ( strcmp( cTemp, "percent of density" ) == 0 )
      p_oNewPackage->SetValue( m_iAmountTypeCode, percentDen );
    else if ( strcmp( cTemp, "absolute basal area" ) == 0 )
      p_oNewPackage->SetValue( m_iAmountTypeCode, absBA );
    else if ( strcmp( cTemp, "absolute density" ) == 0 )
      p_oNewPackage->SetValue( m_iAmountTypeCode, absDen );
    else
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clDisturbance::ReadMortEpParameterFileData" );
      strcpy( stcErr.cMoreInfo, "Unrecognized value in ds_cutAmountType." );
      throw( stcErr );
    }

    //Get the cut ranges
    sVal = XMLString::transcode( "ds_dbhRange" );
    p_oCutSpecificsList = p_oElement->getElementsByTagName( sVal );
    XMLString::release(&sVal);

    iNumChildren = p_oCutSpecificsList->getLength();
    //If there are more ranges than allowed, throw an error
    if ( iNumChildren > m_iNumAllowedCutRanges )
    {
      modelErr stcErr;
      stcErr.iErrorCode = BAD_DATA;
      strcpy( stcErr.cFunction, "clDisturbance::ReadMortEpParameterFileData" );
      strcpy( stcErr.cMoreInfo, "Too many dbh cut ranges." );
      throw( stcErr );
    }
    //Or, if there aren't any and this is a partial cut, throw an error
    if ( 0 == iNumChildren )
    {
      modelErr stcErr;
      stcErr.iErrorCode = DATA_MISSING;
      strcpy( stcErr.cFunction, "clDisturbance::ReadMortEpParameterFileData" );
      strcpy( stcErr.cMoreInfo, "ds_dbhRange" );
      throw( stcErr );
    }

    //Get the minimum dbh in each cut range
    p_fRanges = new float[iNumChildren];
    for ( j = 0; j < iNumChildren; j++ )
    {
      p_oNode = p_oCutSpecificsList->item( j );
      p_oChildElement = ( DOMElement * ) p_oNode;
      FillSingleValue( p_oChildElement, "ds_low", & fTemp, true );
      p_fRanges[j] = fTemp;

      //If the value is negative, throw an error
      if ( fTemp < 0 )
      {
        modelErr stcErr;
        stcErr.iErrorCode = BAD_DATA;
        strcpy( stcErr.cFunction, "clDisturbance::ReadMortEpParameterFileData" );
        strcpy( stcErr.cMoreInfo, "Mortality episode cut range value cannot be negative." );
        throw( stcErr );
      }
    }

    //Sort the values and place them in the package
    fMaxVal = 0;
    for ( j = 0; j < iNumChildren; j++ )
      if ( fMaxVal < p_fRanges[j] )
        fMaxVal = p_fRanges[j];

    fMaxVal += 1;

    for ( j = 0; j < iNumChildren; j++ )
    {
      fTemp = fMaxVal;
      for ( k = 0; k < iNumChildren; k++ )
        if ( p_fRanges[k] < fTemp )
        {
          iMinIndex = k;
          fTemp = p_fRanges[k];
        }

      p_oNode = p_oCutSpecificsList->item( iMinIndex );
      p_oChildElement = ( DOMElement * ) p_oNode;

      p_oNewPackage->SetValue( mp_iRangeMinCodes[j], p_fRanges[iMinIndex] );
      FillSingleValue( p_oChildElement, "ds_high", & fTemp, true );
      p_oNewPackage->SetValue( mp_iRangeMaxCodes[j], fTemp );
      FillSingleValue( p_oChildElement, "ds_amountToCut", & fTemp, true );
      p_oNewPackage->SetValue( mp_iRangeAmountCodes[j], fTemp );

      p_fRanges[iMinIndex] = fMaxVal;
    }

    //Create an ID number - just use the loop counter
    p_oNewPackage->SetValue( m_iMasterIDCode, ( int )i );

    //*******************************************
    //Now:  Get the list of cells to which to apply this mortality episode
    //and make a package for each
    //*******************************************
    sVal = XMLString::transcode( "ds_applyToCell" );
    p_oCutSpecificsList = p_oElement->getElementsByTagName( sVal );
    XMLString::release(&sVal); 
    iNumChildren = p_oCutSpecificsList->getLength();
    for ( j = 0; j < iNumChildren; j++ )
    {
      p_oNode = p_oCutSpecificsList->item( j );
      p_oChildElement = ( DOMElement * ) p_oNode;
      sVal = XMLString::transcode( "x" );
      cData = XMLString::transcode( p_oChildElement->getAttributeNode( sVal )->getNodeValue() );
      iX = atoi( cData );
      delete[] cData; cData = NULL;
      XMLString::release(&sVal);
      sVal = XMLString::transcode( "y" );
      cData = XMLString::transcode( p_oChildElement->getAttributeNode( sVal )->getNodeValue() );
      iY = atoi( cData );
      delete[] cData; cData = NULL;
      XMLString::release(&sVal);

      p_oOldPackage = NULL;
      p_oNewPackage = NULL;
      p_oOldPackage = mp_oCutEventsGrid->GetFirstPackageOfCell( iX, iY );
      while ( p_oOldPackage )
      {
        //Get this package's timestep
        p_oOldPackage->GetValue( m_iCutTimestepCode, & iTemp );

        if ( iTemp >= iTimestep ) //here's where we'll drop it
               break;

        p_oNewPackage = p_oOldPackage; //use temporarily as a holder
        p_oOldPackage = p_oOldPackage->GetNextPackage();
      }

      //Create the new package
      if ( !p_oNewPackage ) //either there's no packages in grid, or new is first
             p_oNewPackage = mp_oCutEventsGrid->CreatePackageOfCell( iX, iY );
      else //make the new one after the one in p_oNewPackage
             p_oNewPackage = mp_oCutEventsGrid->CreatePackage( p_oNewPackage );

      //Load up the new package with the timestep and id values
      p_oNewPackage->SetValue( m_iCutTimestepCode, iTimestep );
      p_oNewPackage->SetValue( m_iCutIDCode, ( int )i );
    } //end of for (j = 0; j < iNumChildren; j++)

    delete[] p_fRanges;
  } //end of for (i = 0; i < iNumDisturbanceEvents; i++)
}



//////////////////////////////////////////////////////////////////////////////
// GetData()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// May 11, 2004 - Split out the parameter file reading stuff (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::GetData( xercesc::DOMDocument * p_oDoc )
{
  try
  {
    clPopulationBase * p_oTemp = mp_oSimManager->GetPopulationObject( "treepopulation" );
    mp_oPop = ( clTreePopulation * ) p_oTemp;

    //Set up the harvest info grid
    SetupGrids();

    //Read the appropriate parameter file data
    if ( m_bIsHarvest )
    {
      ReadHarvestParameterFileData( p_oDoc );
    }
    else
    {
      ReadMortEpParameterFileData( p_oDoc );
    }

    //Validate the package data
    ValidatePackages();

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::GetData" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// ValidatePackages()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// May 11, 2004 - Updated to take episodic mortality into account
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::ValidatePackages()
{
  try
  {
    clPackage * p_oPackage, //a package to validate
         * p_oNextPackage; //for comparing other packages to this one
    float fLowDbh, //low dbh in a cut range
         fHiDbh, //high dbh in a cut range
         fPrevHiDbh, //high dbh in the previous cut range
         fAmount; //cut amount in a cut range
    int iAmountType, //amount type value
         iCutType, //cut type value
         iID, //package ID value
         iNextID; //for comparing other packages' ID values
    short int iNumXCells = mp_oCutEventsGrid->GetNumberXCells(), iNumYCells = mp_oCutEventsGrid->GetNumberYCells(), i, j, k;

    //****************************************************
    // Validate packages in master cuts grid
    //****************************************************
    p_oPackage = mp_oMasterCutsGrid->GetFirstPackageOfCell( 0, 0 );
    while ( p_oPackage )
    {
      fPrevHiDbh = -1;
      if ( m_bIsHarvest )
      {
        p_oPackage->GetValue( m_iCutTypeCode, & iCutType );
      }
      else
      {
        iCutType = partial;
      }

      //Check each cut range to make sure that the high dbh value is bigger
      //than the low value
      for ( k = 0; k < m_iNumAllowedCutRanges; k++ )
      {
        p_oPackage->GetValue( mp_iRangeMinCodes[k], & fLowDbh );
        p_oPackage->GetValue( mp_iRangeMaxCodes[k], & fHiDbh );
        p_oPackage->GetValue( mp_iRangeAmountCodes[k], & fAmount );
        p_oPackage->GetValue( m_iAmountTypeCode, & iAmountType );

        //The range may not have been filled - check
        if ( fLowDbh != 0 && fHiDbh != 0 && fAmount != 0 )
        {

          //Only partial cuts can have more than one cut range defined
          if ( partial != iCutType && k > 0 )
          {
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clDisturbance::ValidatePackages" );
            strcpy( stcErr.cMoreInfo, "Cut ranges may not be defined for harvest gap or clear cuts." );
            throw( stcErr );
          }

          //Make sure the range high dbh value is greater than the low dbh
          if ( fHiDbh <= fLowDbh )
          {
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clDisturbance::ValidatePackages" );
            sprintf( stcErr.cMoreInfo, "%s%s", "The low dbh range value in a", " harvest cut must be smaller than the high." );
            throw( stcErr );
          }

          //Make sure no value is negative
          if ( fLowDbh < 0 || fHiDbh < 0 || fAmount < 0 )
          {
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clDisturbance::ValidatePackages" );
            strcpy( stcErr.cMoreInfo, "No value in a harvest cut range may be negative." );
            throw( stcErr );
          }

          //Make sure that the ranges don't overlap
          if ( fPrevHiDbh > fLowDbh )
          {
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clDisturbance::ValidatePackages" );
            strcpy( stcErr.cMoreInfo, "Harvest cut ranges may not overlap." );
            throw( stcErr );
          }

          //Make sure that the amount value, if a percentage, is not greater
          //than 100
          if ( ( percentBA == iAmountType || percentDen == iAmountType ) && fAmount > 100 )
          {
            modelErr stcErr;
            stcErr.iErrorCode = BAD_DATA;
            strcpy( stcErr.cFunction, "clDisturbance::ValidatePackages" );
            strcpy( stcErr.cMoreInfo, "Harvest cut percentage amounts must be less than 100." );
            throw( stcErr );
          }

          fPrevHiDbh = fHiDbh;
        } //end of if (fLowDbh != 0 && fHiDbh != 0 && fAmount != 0)
      } //end of for (k = 0; k < m_iNumAllowedCutRanges; k++)

      //If this was a gap or clear cut, set up the first cut range for the
      //sake of CutTrees(), if it wasn't passed in.  Set it up to remove all trees
      if ( clear == iCutType || gap == iCutType )
      {

        p_oPackage->GetValue( mp_iRangeAmountCodes[0], & fAmount );
        if ( fAmount == 0 ) fAmount = 100;
        fHiDbh = 3000;
        iAmountType = percentDen;
        p_oPackage->SetValue( mp_iRangeMaxCodes[0], fHiDbh );
        p_oPackage->SetValue( mp_iRangeAmountCodes[0], fAmount );
        p_oPackage->SetValue( m_iAmountTypeCode, iAmountType );
      }
      p_oPackage = p_oPackage->GetNextPackage();
    } //end of while (p_oPackage)

    //****************************************************
    // Validate packages in cut events grid
    //****************************************************
    for ( i = 0; i < iNumXCells; i++ )
      for ( j = 0; j < iNumYCells; j++ )
      {
        p_oPackage = mp_oCutEventsGrid->GetFirstPackageOfCell( i, j );
        while ( p_oPackage )
        {

          //Go through all the other packages in this grid cell and make sure that
          //there is no other with the same ID number
          p_oPackage->GetValue( m_iCutIDCode, & iID );

          p_oNextPackage = p_oPackage->GetNextPackage();
          while ( p_oNextPackage )
          {

            p_oNextPackage->GetValue( m_iCutIDCode, & iNextID );
            if ( iID == iNextID )
            {
              //throw an error
              modelErr stcErr;
              stcErr.iErrorCode = BAD_DATA;
              strcpy( stcErr.cFunction, "clDisturbance::ValidatePackages" );
              strcpy( stcErr.cMoreInfo, "Disturbance event may not be applied twice to the same grid cell." );
              throw( stcErr );
            }
            p_oNextPackage = p_oNextPackage->GetNextPackage();
          } //end of while (p_oNextPackage)
          p_oPackage = p_oPackage->GetNextPackage();
        } //end of while (p_oPackage)
      } //end of for (j = 0; j < iNumYCells; j++)
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::ValidatePackages" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// ResetResultsGrid()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// April 20, 2005 - Changed the name from TimestepCleanup so we could call it
// at a different time (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::ResetResultsGrid()
{
  try
  {
    float fValue = 0;
    int iValue;
    short int iNumXCells = mp_oResultsGrid->GetNumberXCells(), iNumYCells = mp_oResultsGrid->GetNumberYCells(),
         iNumSpecies = mp_oPop->GetNumberOfSpecies(), i, j, iSp, iRange; //loop counters

    //For the harvest results grid - reset all values
    for ( i = 0; i < iNumXCells; i++ )
      for ( j = 0; j < iNumYCells; j++ )
      {

        iValue = -1;

        if ( m_bIsHarvest )
        {
          //Harvest Type - set values to -1 (no harvest event has occurred)
          mp_oResultsGrid->SetValueOfCell( i, j, m_iHarvestTypeCode, iValue );
        }

        iValue = 0;
        //Cut Density and Cut Basal Area - set values to 0
        for ( iSp = 0; iSp < iNumSpecies; iSp++ )
          for ( iRange = 0; iRange < m_iNumAllowedCutRanges; iRange++ )
          {
            mp_oResultsGrid->SetValueOfCell( i, j, mp_iDenCutCodes[iRange] [iSp], iValue );
            mp_oResultsGrid->SetValueOfCell( i, j, mp_iBaCutCodes[iRange] [iSp], fValue );
          }
      }

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::ResetResultsGrid" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// SetupGrids()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// May 11, 2004 - Added support for episodic mortality grids (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::SetupGrids()
{
  try
  {
    clPlot * p_oPlot = mp_oSimManager->GetPlotObject();
    char cFloatLabel[MAX_GRID_LABEL_SIZE];
    float fCellLength = mp_oPop->GetGridCellSize();
    short int iNumSpecies = mp_oPop->GetNumberOfSpecies(), i, j; //loop counters

    if ( m_bIsHarvest )
    {
      //*************************************
      //harvestmastercuts grid
      //*************************************
      mp_oMasterCutsGrid = mp_oSimManager->CreateGrid( "harvestmastercuts", //grid name
           0, //number of int data members
           0, //number of float data members
           0, //number of char data members
           0, //number of bool data members
           p_oPlot->GetXPlotLength(), //X cell length
           p_oPlot->GetYPlotLength() ); //Y cell length

      //Now adjust for the package data members
      mp_oMasterCutsGrid->ChangePackageDataStructure( 4, //number of package int data members
           ( 3 * m_iNumAllowedCutRanges ), //number of package float data members
           0, //number of package char data members
           iNumSpecies ); //number of package bool data members

      //Declare the arrays for holding data member codes
      mp_iRangeMinCodes = new short int[m_iNumAllowedCutRanges];
      mp_iRangeMaxCodes = new short int[m_iNumAllowedCutRanges];
      mp_iRangeAmountCodes = new short int[m_iNumAllowedCutRanges];
      mp_iDenCutCodes = new short int * [m_iNumAllowedCutRanges];
      mp_iBaCutCodes = new short int * [m_iNumAllowedCutRanges];
      for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
      {
        mp_iDenCutCodes[i] = new short int[iNumSpecies];
        mp_iBaCutCodes[i] = new short int[iNumSpecies];
      }
      mp_iSpeciesCodes = new short int[iNumSpecies];

      //Register the data members
      m_iMasterTimestepCode = mp_oMasterCutsGrid->RegisterPackageInt( "timestep" );
      m_iMasterIDCode = mp_oMasterCutsGrid->RegisterPackageInt( "id" );
      m_iCutTypeCode = mp_oMasterCutsGrid->RegisterPackageInt( "cuttype" );
      m_iAmountTypeCode = mp_oMasterCutsGrid->RegisterPackageInt( "amttype" );

      for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
      {
        sprintf( cFloatLabel, "%s%d", "rangemin", i );
        mp_iRangeMinCodes[i] = mp_oMasterCutsGrid->RegisterPackageFloat( cFloatLabel );
        sprintf( cFloatLabel, "%s%d", "rangemax", i );
        mp_iRangeMaxCodes[i] = mp_oMasterCutsGrid->RegisterPackageFloat( cFloatLabel );
        sprintf( cFloatLabel, "%s%d", "rangeamt", i );
        mp_iRangeAmountCodes[i] = mp_oMasterCutsGrid->RegisterPackageFloat( cFloatLabel );
      }

      for ( i = 0; i < iNumSpecies; i++ )
      {
        sprintf( cFloatLabel, "%s%d", "species", i );
        mp_iSpeciesCodes[i] = mp_oMasterCutsGrid->RegisterPackageBool( cFloatLabel );
      }


      //*************************************
      //harvestcutevents grid
      //*************************************
      mp_oCutEventsGrid = mp_oSimManager->CreateGrid( "harvestcutevents", //grid name
           0, //number of int data members
           0, //number of float data members
           0, //number of char data members
           0, //number of bool data members
           fCellLength, //X cell length
           fCellLength ); //Y cell length

      //Now adjust for the package data members
      mp_oCutEventsGrid->ChangePackageDataStructure( 2, //number of package int data members
           0, //number of package float data members
           0, //number of package char data members
           0 ); //number of package bool data members

      //Register the data members
      m_iCutTimestepCode = mp_oCutEventsGrid->RegisterPackageInt( "timestep" );
      m_iCutIDCode = mp_oCutEventsGrid->RegisterPackageInt( "id" );



      //*************************************
      //Harvest results grid
      //*************************************
      mp_oResultsGrid = mp_oSimManager->CreateGrid( "Harvest Results", //grid name
           1 + ( m_iNumAllowedCutRanges * iNumSpecies ), //number of int data members
           ( m_iNumAllowedCutRanges * iNumSpecies ), //num float data members
           0, //number of char data members
           0, //number of bool data members
           fCellLength, //X cell length
           fCellLength ); //Y cell length

      //Register the data members
      m_iHarvestTypeCode = mp_oResultsGrid->RegisterInt( "Harvest Type" );

      for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
        for ( j = 0; j < iNumSpecies; j++ )
        {
          sprintf( cFloatLabel, "%s%d%s%d", "Cut Density_", i, "_", j );
          mp_iDenCutCodes[i] [j] = mp_oResultsGrid->RegisterInt( cFloatLabel );
          sprintf( cFloatLabel, "%s%d%s%d", "Cut Basal Area_", i, "_", j );
          mp_iBaCutCodes[i] [j] = mp_oResultsGrid->RegisterFloat( cFloatLabel );
        }
    }
    else
    {

      //*************************************
      //mortepisodemastercuts grid
      //*************************************
      mp_oMasterCutsGrid = mp_oSimManager->CreateGrid( "mortepisodemastercuts", //grid name
           0, //number of int data members
           0, //number of float data members
           0, //number of char data members
           0, //number of bool data members
           p_oPlot->GetXPlotLength(), //X cell length
           p_oPlot->GetYPlotLength() ); //Y cell length

      //Now adjust for the package data members
      mp_oMasterCutsGrid->ChangePackageDataStructure( 3, //number of package int data members
           ( 3 * m_iNumAllowedCutRanges ), //number of package float data members
           0, //number of package char data members
           iNumSpecies ); //number of package bool data members

      //Declare the arrays for holding data member codes
      mp_iRangeMinCodes = new short int[m_iNumAllowedCutRanges];
      mp_iRangeMaxCodes = new short int[m_iNumAllowedCutRanges];
      mp_iRangeAmountCodes = new short int[m_iNumAllowedCutRanges];
      mp_iDenCutCodes = new short int * [m_iNumAllowedCutRanges];
      mp_iBaCutCodes = new short int * [m_iNumAllowedCutRanges];
      for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
      {
        mp_iDenCutCodes[i] = new short int[iNumSpecies];
        mp_iBaCutCodes[i] = new short int[iNumSpecies];
      }
      mp_iSpeciesCodes = new short int[iNumSpecies];

      //Register the data members
      m_iMasterTimestepCode = mp_oMasterCutsGrid->RegisterPackageInt( "timestep" );
      m_iMasterIDCode = mp_oMasterCutsGrid->RegisterPackageInt( "id" );
      m_iAmountTypeCode = mp_oMasterCutsGrid->RegisterPackageInt( "amttype" );

      for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
      {
        sprintf( cFloatLabel, "%s%d", "rangemin", i );
        mp_iRangeMinCodes[i] = mp_oMasterCutsGrid->RegisterPackageFloat( cFloatLabel );
        sprintf( cFloatLabel, "%s%d", "rangemax", i );
        mp_iRangeMaxCodes[i] = mp_oMasterCutsGrid->RegisterPackageFloat( cFloatLabel );
        sprintf( cFloatLabel, "%s%d", "rangeamt", i );
        mp_iRangeAmountCodes[i] = mp_oMasterCutsGrid->RegisterPackageFloat( cFloatLabel );
      }

      for ( i = 0; i < iNumSpecies; i++ )
      {
        sprintf( cFloatLabel, "%s%d", "species", i );
        mp_iSpeciesCodes[i] = mp_oMasterCutsGrid->RegisterPackageBool( cFloatLabel );
      }


      //*************************************
      //mortepisodecutevents grid
      //*************************************
      mp_oCutEventsGrid = mp_oSimManager->CreateGrid( "mortepisodecutevents", //grid name
           0, //number of int data members
           0, //number of float data members
           0, //number of char data members
           0, //number of bool data members
           fCellLength, //X cell length
           fCellLength ); //Y cell length

      //Now adjust for the package data members
      mp_oCutEventsGrid->ChangePackageDataStructure( 2, //number of package int data members
           0, //number of package float data members
           0, //number of package char data members
           0 ); //number of package bool data members

      //Register the data members
      m_iCutTimestepCode = mp_oCutEventsGrid->RegisterPackageInt( "timestep" );
      m_iCutIDCode = mp_oCutEventsGrid->RegisterPackageInt( "id" );



      //*************************************
      //Mortality Episode Results grid
      //*************************************
      mp_oResultsGrid = mp_oSimManager->CreateGrid( "Mortality Episode Results", //grid name
           ( m_iNumAllowedCutRanges * iNumSpecies ), //number of int data members
           ( m_iNumAllowedCutRanges * iNumSpecies ), //num float data members
           0, //number of char data members
           0, //number of bool data members
           fCellLength, //X cell length
           fCellLength ); //Y cell length

      //Register the data members
      for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
        for ( j = 0; j < iNumSpecies; j++ )
        {
          sprintf( cFloatLabel, "%s%d%s%d", "Cut Density_", i, "_", j );
          mp_iDenCutCodes[i] [j] = mp_oResultsGrid->RegisterInt( cFloatLabel );
          sprintf( cFloatLabel, "%s%d%s%d", "Cut Basal Area_", i, "_", j );
          mp_iBaCutCodes[i] [j] = mp_oResultsGrid->RegisterFloat( cFloatLabel );
        }
    }

    //Use the ResetResultsGrid function to initialize the values
    ResetResultsGrid();

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::SetupHarvestGrid" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// Action()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// April 20, 2005 - Added a call to reset the results grid (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::Action()
{
  try
  {
    clPackage * p_oMasterPackage, * p_oPackageHolder;
    int iTimestep = mp_oSimManager->GetCurrentTimestep(), iPackageTimestep;

    ResetResultsGrid();

    //Get all the master cut packages for this timestep and execute them
    p_oMasterPackage = mp_oMasterCutsGrid->GetFirstPackageOfCell( 0, 0 );

    while ( p_oMasterPackage )
    {
      //Get the next package - if this is a cut package, it will be deleted during
      //the course of the cut process
      p_oPackageHolder = p_oMasterPackage->GetNextPackage();

      //Get this package's timestep
      p_oMasterPackage->GetValue( m_iMasterTimestepCode, & iPackageTimestep );
      if ( iTimestep == iPackageTimestep )
        CutTrees( p_oMasterPackage );

      //I commented this out as a quick-and-dirty way to make this work if we
      //start on a timestep other than 0 and there are defined events that
      //were supposed to happen before the start timestep.
      //      else
      //        break; //we've done them all this timestep

      p_oMasterPackage = p_oPackageHolder;
    }
  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::Action" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// CutTrees()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// April 13, 2005 - Ommitted snags (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::CutTrees( clPackage * p_oMasterPackage )
{
  try
  {
    stcGridList * p_cutArea = NULL, //pointer to list of applicable grid cells
         * p_cellRecord = NULL; //for working with cell records
    clTree * p_oTree, //tree we're working with
         * p_oNextTree; //placeholder for getting the next tree
    float * p_fLoDbh, //low dbh of cut ranges - sized m_iNumAllowedCutRanges
         * p_fHiDbh, //high dbh of cut ranges - sized m_iNumAllowedCutRanges
         * p_fAmountToRemove, //amount to cut - sized m_iNumAllowedCutRanges
         * p_fAmountRemoved, //amount cut - either in basal area or density -
         //sized number of cut ranges
         * p_fBABackup = NULL, //this is to preserve the amt basal area to remove
         * p_fTotalBasalArea = NULL, //amount of total basal area - sized number
         //of cut ranges
         fRand, //random number
         fDbh, //dbh of tree
         fThisBA, //basal area of a single tree
         fTemp, //for doing grid values
         fX, fY; //tree coordinates
    int iNumXCells = mp_oCutEventsGrid->GetNumberXCells(), iNumYCells = mp_oCutEventsGrid->GetNumberYCells(),
         iNumSpecies = mp_oPop->GetNumberOfSpecies(), iAmountType, //in what terms the cut is expressed
         iCutType, //type of cut to be performed
         iTemp, //for doing grid values
         iNumCellsInCutArea; //number of cells in cut area
    short int iSp, //species of tree
         iDeadCode, //for testing if a tree is dead
         i; //loop counter
    bool * p_bSpeciesCut, //array of size # species of whether each species is in
         //this cut
         bTreeIsDead; //whether or not a particular tree dies



    //#define test;
#ifdef test
    float fTemp2;
    fstream harv( "HarvestTrees.xls", ios::out | ios::app );
    harv << "Timestep: " << mp_oSimManager->GetCurrentTimestep() << "\n" << "X\tY\tSpecies\tDbh\n";
#endif



    //Declare our arrays and set all the values to 0
    p_fLoDbh = new float[m_iNumAllowedCutRanges];
    p_fHiDbh = new float[m_iNumAllowedCutRanges];
    p_fAmountToRemove = new float[m_iNumAllowedCutRanges];
    p_fAmountRemoved = new float[m_iNumAllowedCutRanges];
    for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
    {
      p_fLoDbh[i] = 0;
      p_fHiDbh[i] = 0;
      p_fAmountToRemove[i] = 0;
      //p_fAmountRemoved[i] = 0; no need - we do it later
    }

    p_bSpeciesCut = new bool[iNumSpecies];
    for ( i = 0; i < iNumSpecies; i++ )
      p_bSpeciesCut[i] = false;

    //Get the cut type and cut amount values for this package
    if ( m_bIsHarvest )
    {
      p_oMasterPackage->GetValue( m_iCutTypeCode, & iCutType );
    }
    else
    {
      iCutType = partial;
    }
    p_oMasterPackage->GetValue( m_iAmountTypeCode, & iAmountType );

    //If the cut type is percent of basal area - declare the total basal area
    //arrays
    if ( partial == iCutType && percentBA == iAmountType )
    {
      p_fTotalBasalArea = new float[m_iNumAllowedCutRanges];
      p_fBABackup = new float[m_iNumAllowedCutRanges];
    }

    //Get the cut area data - the cut information extracted from the master
    //package and the linked list of grid cells to which to apply the cut
    iNumCellsInCutArea = AssembleCutArea( p_oMasterPackage, iNumXCells, iNumYCells, p_cutArea, p_fLoDbh, p_fHiDbh,
         p_fAmountToRemove, p_bSpeciesCut );

    //Set the cut event that's occurring for each cell in the cut area in the
    //harvest results grid
    if ( m_bIsHarvest ) SetCutFlags( p_cutArea, iCutType );

    //Convert the amount to cut - if it's in percentage, convert it to a
    //proportion; if it's an absolute value in amounts per hectare, scale it to
    //a total amount
    if ( percentDen == iAmountType || percentBA == iAmountType ) //percentage value
           for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
             p_fAmountToRemove[i] *= 0.01;
    else if ( absDen == iAmountType || absBA == iAmountType )
    {
      //Figure out the area of the cut in hectares and multiply
      fTemp = iNumCellsInCutArea * mp_oCutEventsGrid->GetLengthXCells() * mp_oCutEventsGrid->GetLengthYCells() / 10000;
      for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
        p_fAmountToRemove[i] *= fTemp;
    }
    if ( absDen == iAmountType ) //Round to nearest whole number
           for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
             p_fAmountToRemove[i] = clModelMath::Round( p_fAmountToRemove[i], 0 );

    //Cut each species in turn for this package
    for ( iSp = 0; iSp < iNumSpecies; iSp++ )
      if ( p_bSpeciesCut[iSp] )
      {

        //Reset the amount removed array
        for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
          p_fAmountRemoved[i] = 0;

        //If the cut type is percentage of basal area, we have to know how
        //much basal area is in the cut ranges.  So get it.
        if ( percentBA == iAmountType )
        {

          //Zero out the basal area array
          for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
            p_fTotalBasalArea[i] = 0;

          GetBasalArea( p_cutArea, iSp, p_fTotalBasalArea, p_fLoDbh, p_fHiDbh );

          //Back up the amount value
          for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
            p_fBABackup[i] = p_fAmountToRemove[i];

          //Multiply the percentage to cut by the total basal area and then treat
          //the cut like an amount of absolute basal area
          for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
            p_fAmountToRemove[i] *= p_fTotalBasalArea[i];
        }


        //*******************************************
        //Now remove the trees
        //*******************************************
        p_oTree = GetTallestTreeInCutArea( p_cutArea, iSp );
        while ( p_oTree )
        {
          bTreeIsDead = false;
          fThisBA = 0;

          //See if this tree already died in another SORTIE disturbance
          iDeadCode = mp_oPop->GetBoolDataCode( "dead", p_oTree->GetSpecies(), p_oTree->GetType() );
          if ( -1 != iDeadCode )
          {
            p_oTree->GetValue( iDeadCode, & bTreeIsDead );
          }
          else
            bTreeIsDead = false;

          if ( !bTreeIsDead && p_oTree->GetType() != clTreePopulation::snag )
          {

            p_oTree->GetValue( mp_oPop->GetDbhCode( iSp, p_oTree->GetType() ), & fDbh );

            //Go through each cut range and see if a tree is in it
            for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
            {
              if ( fDbh >= p_fLoDbh[i] && fDbh <= p_fHiDbh[i] )
              {

                //What cut amount type are we talking about here?

                //Percent of density - compare a random number to the desired
                //percentage to remove
                if ( percentDen == iAmountType )
                {
                  fRand = clModelMath::GetRand();
                  if ( fRand <= p_fAmountToRemove[i] )
                    bTreeIsDead = true;




                  //Absolute density - keep cutting until we've cut down all that
                  //we're supposed to
                }
                else if ( absDen == iAmountType )
                {
                  if ( p_fAmountRemoved[i] < p_fAmountToRemove[i] )
                  {
                    bTreeIsDead = true;
                    p_fAmountRemoved[i] += 1.0;
                  }
                  else
                    p_oTree = NULL; //break the loop




                  //Absolute amount of basal area (and percent of basal area convert-
                  //ed to absolute) - cut until we've reached the goal
                }
                else if ( absBA == iAmountType || percentBA == iAmountType )
                {
                  if ( p_fAmountRemoved[i] < p_fAmountToRemove[i] )
                  {
                    fThisBA = clModelMath::CalculateBasalArea( fDbh );
                    p_fAmountRemoved[i] += fThisBA;
                    //Don't go over the limit - maybe a smaller tree will get us
                    //closer
                    if ( p_fAmountRemoved[i] <= p_fAmountToRemove[i] )
                      bTreeIsDead = true;
                    else
                      p_fAmountRemoved[i] -= fThisBA;
                  } //end of if (p_fAmountRemoved[i] < p_fAmountToRemove[i])
                } //end of else if (absBA == iAmountType || percentBA == ...)
              } //end of (fDbh >= p_fLoDbh[i] && fDbh <= p_fHiDbh[i])
            } //end of for (i = 0; i < m_iNumAllowedCutRanges; i++)
          }
          else
          {
            //tree was dead or a snag - but set it to NOT dead so that
            //it doesn't get cut
            bTreeIsDead = false;
          }

          p_oNextTree = GetNextTreeInCutArea( p_cutArea, iSp );
          if ( bTreeIsDead )
          {

            //Add the dead tree's stats to the grid keeping track of harvest cuts
            p_oTree->GetValue( mp_oPop->GetXCode( iSp, p_oTree->GetType() ), & fX );
            p_oTree->GetValue( mp_oPop->GetYCode( iSp, p_oTree->GetType() ), & fY );

            for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
              if ( fDbh >= p_fLoDbh[i] && fDbh <= p_fHiDbh[i] )
              {
                mp_oResultsGrid->GetValueAtPoint( fX, fY, mp_iDenCutCodes[i] [iSp], & iTemp );
                iTemp += 1;
                mp_oResultsGrid->SetValueAtPoint( fX, fY, mp_iDenCutCodes[i] [iSp], iTemp );

                if ( 0 == fThisBA )
                  fThisBA = clModelMath::CalculateBasalArea( fDbh );

                mp_oResultsGrid->GetValueAtPoint( fX, fY, mp_iBaCutCodes[i] [iSp], & fTemp );
                fTemp += fThisBA;
                mp_oResultsGrid->SetValueAtPoint( fX, fY, mp_iBaCutCodes[i] [iSp], fTemp );
              } //end of if (fDbh >= p_fLoDbh[i] && fDbh <= p_fHiDbh[i])


#ifdef test
            //Get dbh
            p_oTree->GetValue( mp_oPop->GetDbhCode( iSp, p_oTree->GetType() ), & fTemp2 );
            harv << fX << "\t" << fY << "\t" << iSp << "\t" << fTemp2 << "\n";
#endif

            mp_oPop->KillTree( p_oTree, m_iReasonCode );
          }

          p_oTree = p_oNextTree;
        } //end of while (p_oTree)

        //If this was percent of basal area, reset the amounts to the backup vals
        if ( percentBA == iAmountType )
          for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
            p_fAmountToRemove[i] = p_fBABackup[i];
      } //end of if (p_bSpeciesCut[iSp])

    delete[] p_fLoDbh;
    delete[] p_fHiDbh;
    delete[] p_fAmountToRemove;
    delete[] p_fAmountRemoved;
    delete[] p_fTotalBasalArea;
    delete[] p_bSpeciesCut;
    delete[] p_fBABackup;

    //Delete the list of grid cell records in this cut area
    while ( p_cutArea )
    {
      p_cellRecord = p_cutArea->next;
      delete p_cutArea;
      p_cutArea = p_cellRecord;
    }


#ifdef test
    harv.close();
#endif


  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::CutTrees" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// AssembleCutArea()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
int clDisturbance::AssembleCutArea( clPackage * p_oMasterPackage, const int & iNumXCells, const int & iNumYCells,
     stcGridList * & p_cutArea, float * p_fLoDbh, float * p_fHiDbh, float * p_fAmountToRemove, bool * p_bSpeciesCut )
     {
       try
       {
         struct stcGridList * p_cellRecord = NULL; //for creating cut area
         float fTemp; //for extracting package values
         int iMasterID, //ID of master cell
              iTemp, //for getting package integer values
              iNumSpecies = mp_oPop->GetNumberOfSpecies(), iTimestep = mp_oSimManager->GetCurrentTimestep(), iNumCellsInCutArea = 0; //number of
         // cells in the cut area
         short int i, j; //loop counters
         bool bTemp; //for getting package bool values

         //********************************************
         // Extract the data from the master package
         //********************************************

         //Get the master package's ID number
         p_oMasterPackage->GetValue( m_iMasterIDCode, & iMasterID );

         //Get the package's cut ranges and amounts
         for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
         {
           //Low dbh
           p_oMasterPackage->GetValue( mp_iRangeMinCodes[i], & fTemp );
           p_fLoDbh[i] = fTemp;
           //High dbh
           p_oMasterPackage->GetValue( mp_iRangeMaxCodes[i], & fTemp );
           p_fHiDbh[i] = fTemp;
           //Amount
           p_oMasterPackage->GetValue( mp_iRangeAmountCodes[i], & fTemp );
           p_fAmountToRemove[i] = fTemp;
         }

         //Get the species list
         for ( i = 0; i < iNumSpecies; i++ )
         {
           p_oMasterPackage->GetValue( mp_iSpeciesCodes[i], & bTemp );
           p_bSpeciesCut[i] = bTemp;
         }

         //Delete the master package - we're done with it
         mp_oMasterCutsGrid->DeletePackage( p_oMasterPackage );


         //********************************************
         // Find the matching cut packages and assemble them into
         // our linked list
         //********************************************

         for ( i = 0; i < iNumXCells; i++ )
           for ( j = 0; j < iNumYCells; j++ )
           {

             //Reusing the p_oMasterPackage pointer
             p_oMasterPackage = mp_oCutEventsGrid->GetFirstPackageOfCell( i, j );
             while ( p_oMasterPackage )
             {

               //Get this package's timestep to see if it's for this timestep
               p_oMasterPackage->GetValue( m_iCutTimestepCode, & iTemp );
               if ( iTimestep != iTemp )
               {
                 //Null out the pointer - this will break the loop
                 p_oMasterPackage = NULL;
                 goto nextPackage;
               }

               //Check the ID number
               p_oMasterPackage->GetValue( m_iCutIDCode, & iTemp );
               if ( iTemp != iMasterID ) goto nextPackage;

               //Make this a record in our cut list - stick it at the beginning
               if ( p_cutArea != NULL )
               {
                 p_cellRecord = new stcGridList;
                 p_cellRecord->iX = i;
                 p_cellRecord->iY = j;
                 p_cellRecord->p_oTree = NULL;
                 p_cellRecord->fDbh = 0;
                 p_cellRecord->next = p_cutArea;
                 p_cutArea = p_cellRecord;
               }
               else
               {
                 //Provision for first cut in linked list
                 p_cutArea = new stcGridList;
                 p_cutArea->iX = i;
                 p_cutArea->iY = j;
                 p_cutArea->next = NULL;
                 p_cutArea->p_oTree = NULL;
                 p_cutArea->fDbh = 0;
               }

               iNumCellsInCutArea++;

               //We're done with this package - delete it
               mp_oCutEventsGrid->DeletePackage( p_oMasterPackage );
               p_oMasterPackage = NULL;

             nextPackage:
               if ( p_oMasterPackage ) p_oMasterPackage = p_oMasterPackage->GetNextPackage();
             } //end of while (p_oMasterPackage)
           } //end of for (j = 0; j < iNumYCells; j++)

         return iNumCellsInCutArea;
       } //end of try block
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clDisturbance::AssembleCutArea" );
         throw( stcErr );
       }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// GetBasalArea()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
// April 13, 2005 - Ommitted snags (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::GetBasalArea( stcGridList * p_cutArea, const short int & iSpecies, float * p_fTotalBasalArea,
     float * p_fLoDbh, float * p_fHiDbh )
     {
       try
       {
         clTree * p_oTree; //tree whose basal area we're getting
         stcGridList * p_cellRecord = NULL; //grid cell
         float fDbh; //tree's dbh
         short int iTreeType, iTreeSpecies, //tree's type and species
         i; //loop counter

         p_cellRecord = p_cutArea;
         while ( p_cellRecord )
         {
           p_oTree = mp_oPop->GetTallestTreeInCell( p_cellRecord->iX, p_cellRecord->iY );
           while ( p_oTree )
           {
             iTreeSpecies = p_oTree->GetSpecies();
             if ( iSpecies == iTreeSpecies )
             {
               iTreeType = p_oTree->GetType();
               if ( clTreePopulation::seedling != iTreeType && clTreePopulation::snag != iTreeType )
               {
                 p_oTree->GetValue( mp_oPop->GetDbhCode( iTreeSpecies, iTreeType ), & fDbh );
                 for ( i = 0; i < m_iNumAllowedCutRanges; i++ )
                   if ( fDbh >= p_fLoDbh[i] && fDbh <= p_fHiDbh[i] )
                     p_fTotalBasalArea[i] += clModelMath::CalculateBasalArea( fDbh );
               }
             } //end of if (iSpecies == iTreeSpecies)
             p_oTree = p_oTree->GetShorter();
           } //end of while (p_oTree)
           p_cellRecord = p_cellRecord->next;
         } //end of while (p_cellRecord)

       } //end of try block
       catch ( modelErr & err )
       {
         throw( err );
       }
       catch ( modelMsg & msg )
       {
         throw( msg );
       } //non-fatal error
       catch ( ... )
       {
         modelErr stcErr;
         stcErr.iErrorCode = UNKNOWN;
         strcpy( stcErr.cFunction, "clDisturbance::GetBasalArea" );
         throw( stcErr );
       }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// GetTallestTreeInCutArea()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
clTree * clDisturbance::GetTallestTreeInCutArea( stcGridList * & p_cutArea, const short int & iSpecies )
{
  try
  {
    clTree * p_oTree; //a tree we're working with
    stcGridList * p_sortedCutArea = NULL, //where to put the sorted list
    * p_cellRecord, //one record in the cut area list
    * p_cellHolder, //for holding our place in the linked list
    * p_sorter; //for sorting the linked list
    float fDbh; //tree's dbh

    //Go through each of the cell list records and find the tallest tree of
    //the given species
    p_cellRecord = p_cutArea;
    while ( p_cellRecord )
    {
      p_cellHolder = p_cellRecord->next;
      p_cellRecord->p_oTree = NULL;
      p_cellRecord->fDbh = 0;
      p_cellRecord->next = NULL;
      fDbh = 0;

      //Go through the trees to find the tallest
      p_oTree = mp_oPop->GetTallestTreeInCell( p_cellRecord->iX, p_cellRecord->iY );
      while ( p_oTree )
      {
        if ( p_oTree->GetSpecies() == iSpecies )
        {
          if ( clTreePopulation::seedling == p_oTree->GetType() )
          {
            //No more trees to bother with here - null the pointer and break
            //the loop
            p_oTree = NULL;
            break;
          }
          else
          {
            break;
          }
        }
        p_oTree = p_oTree->GetShorter();
      } //end of while (p_oTree)

      //Get the dbh of our tree, if it exists
      if ( p_oTree )
        p_oTree->GetValue( mp_oPop->GetDbhCode( iSpecies, p_oTree->GetType() ), & fDbh );
      else
        fDbh = 0;

      //Assign the values
      p_cellRecord->p_oTree = p_oTree;
      p_cellRecord->fDbh = fDbh;

      //Sort our record (tallest tree or not) into the new list
      p_sorter = p_sortedCutArea;
      if ( NULL != p_sorter )
      {
        while ( p_sorter )
        {

          //Does our record come first?
          if ( p_sorter == p_sortedCutArea && p_sorter->fDbh <= fDbh )
          {
            p_cellRecord->next = p_sorter;
            p_sortedCutArea = p_cellRecord;
            break;
          }

          //Evaluate the other cases independently of the first - no else if here
          //Does our record come last?
          if ( NULL == p_sorter->next )
          {
            p_sorter->next = p_cellRecord;
            break;

            //Our record comes somewhere in the middle
          }
          else if ( p_sorter->next->fDbh <= fDbh )
          {

            //The new record goes after p_sorter
            p_cellRecord->next = p_sorter->next;
            p_sorter->next = p_cellRecord;
            break;
          }

          p_sorter = p_sorter->next;
        } //end of while (p_sorter)
      }
      else //this is first one to be sorted
        p_sortedCutArea = p_cellRecord;

      p_cellRecord = p_cellHolder;
    } //end of while (p_cellRecord)

    //OK, we're all sorted - return the tree in the first block and re-assign
    //the linked list
    p_cutArea = p_sortedCutArea;
    return ( p_cutArea->p_oTree );

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::GetTallestTreeInCutArea" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// GetNextTreeInCutArea()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
clTree * clDisturbance::GetNextTreeInCutArea( stcGridList * & p_cutArea, const short int & iSpecies )
{
  try
  {
    clTree * p_oTree; //for finding the next tallest tree
    stcGridList * p_cellRecord, //an individual record
    * p_sorter; //for sorting the list
    float fDbh; //tree's dbh

    //If we've already returned all the trees return NULL
    if ( NULL == p_cutArea->p_oTree ) return NULL;

    //The tallest tree is the one in the first cut record, and it's the one
    //that's no longer wanted.  So update that cell with the next tallest one
    //of that species
    p_oTree = p_cutArea->p_oTree->GetShorter();
    while ( p_oTree )
    {
      if ( p_oTree->GetSpecies() == iSpecies )
        if ( clTreePopulation::seedling != p_oTree->GetType() )
          break;
        else
        {
          p_oTree = NULL;
          break;
        }
      p_oTree = p_oTree->GetShorter();
    } //end of while (p_oTree)

    p_cellRecord = p_cutArea;
    if ( p_oTree )
      p_oTree->GetValue( mp_oPop->GetDbhCode( iSpecies, p_oTree->GetType() ), & fDbh );
    else
      fDbh = 0;

    p_cellRecord->p_oTree = p_oTree;
    p_cellRecord->fDbh = fDbh;

    //Now move this record down in the list if necessary
    if ( p_cutArea->next && fDbh < p_cutArea->next->fDbh )
    { //yep - needs to move
      p_sorter = p_cutArea->next;
      p_cutArea = p_sorter;
      while ( p_sorter )
      {
        //Does our record come last?
        if ( NULL == p_sorter->next )
        {
          p_sorter->next = p_cellRecord;
          p_cellRecord->next = NULL;
          break;

          //Our record comes somewhere in the middle
        }
        else if ( p_sorter->next->fDbh <= fDbh )
        {

          //The new record goes after p_sorter
          p_cellRecord->next = p_sorter->next;
          p_sorter->next = p_cellRecord;
          break;
        }

        p_sorter = p_sorter->next;
      } //end of while (p_sorter)
    } //end of if (fDbh < p_oCutArea->next->fDbh)

    return ( p_cutArea->p_oTree );

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::GetNextTreeInCutArea" );
    throw( stcErr );
  }
}

//----------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
// SetCutFlags()
// Edit history:
// -------------
// April 28, 2004 - Submitted in beta version (LEM)
//////////////////////////////////////////////////////////////////////////////
void clDisturbance::SetCutFlags( stcGridList * p_cutArea, const int & iCutType )
{
  try
  {
    stcGridList * p_record;
    int iTemp;

    p_record = p_cutArea;
    while ( p_record )
    {

      //Make sure the cut severity type is set to the most severe type for
      //this grid cell
      mp_oResultsGrid->GetValueOfCell( p_record->iX, p_record->iY, m_iHarvestTypeCode, & iTemp );

      if ( iTemp < iCutType )
        mp_oResultsGrid->SetValueOfCell( p_record->iX, p_record->iY, m_iHarvestTypeCode, iCutType );

      p_record = p_record->next;
    }

  } //end of try block
  catch ( modelErr & err )
  {
    throw( err );
  }
  catch ( modelMsg & msg )
  {
    throw( msg );
  } //non-fatal error
  catch ( ... )
  {
    modelErr stcErr;
    stcErr.iErrorCode = UNKNOWN;
    strcpy( stcErr.cFunction, "clDisturbance::SetCutFlags" );
    throw( stcErr );
  }
}
//----------------------------------------------------------------------------
